// MFile.java

/*
This software is part of the Gidlet framework.
It is Open Source Free Software, so you may
- run the code for any purpose
- study how the code works and adapt it to your needs
- integrate all or parts of the code in your own programs
- redistribute copies of the code
- improve the code and release your improvements to the public
However the use of the code is entirely your responsibility.
 */

package ch.aplu.gidlet;

import javax.microedition.io.*;
import javax.microedition.io.file.*;
import java.io.*;
import java.util.*;

/**
 * Class to access files or directories that are located on removable media
 * and/or file systems on a device.
 * Integrates the the file connection package (JSR 75) into the Gidlet framework,
 * in particular errors are not reported by exceptions but by boolean return values.
 * Not all mobile devices support JSR 75. Use isFileConnectionSupported()
 * for dynamic detection of support or consult the manufacture technical specifations.
 * File read and write operations are performed in a separate thread.
 * For all methods, except getDataInputStream(), getDataOutputStream(), openConnection().
 * the following schema is used:<br>
 * - close an already opened file connection<br>
 * - open a new file connection<br>
 * - perform operation<br>
 * - close the file connection<br<br>
 *
 * To suppress the question that the emulator ask about accessing the file system
 * change in file<br>
 * - c:\Documents and Settings\<user>\j2mewtk\2.5.2\appdb\_policy.txt<br>
 *   (for WTK 2.5.2 and NetBeans 5.5.1 up)<br><br>
 *
 * - wtkroot\appdb\_policy.txt<br>
 *   (for WTK 2.5.1)<br><br>
 *
 * existing lines<br><br>
 *
 * blanket(oneshot): read_user_data_access<br>
 * blanket(oneshot): write_user_data_access<br><br>
 *
 * to<br><br>
 *
 * allow: read_user_data_access<br>
 * allow: write_user_data_access<br><br><br>
 *
 *
 * When using the WTK emulator with the DefaultColorPhone the home directory
 * of the emulated file system resides on<br>
 * - userhome\j2mewtk\2.5.2\appdb\DefaultColorPhone\filesystem<br>
 *   (for WTK 2.5.2 and NetBeans 5.5.1 up)<br><br>
 *
 * - wtkroot\appdb\DefaultColorPhone\filesystem<br>
 *   (for WTK 2.5.1)<br>
 *
 * Only one file system root (emulated device) is visible (name: root1/).
 */
public class MFile
{

// --------------------- Inner class ReadThread ------------------
  private class ReadThread extends Thread
  {

    private StringBuffer sb;

    ReadThread(StringBuffer sb)
    {
      this.sb = sb;
    }

    public void run()
    {
      try
      {
        openfc();
      }
      catch (IOException ex)
      {
        isError = true;
        return;
      }

      InputStream is = null;

      try
      {
        is = fc.openInputStream();

        int ch;
        while ((ch = is.read()) != -1)
          sb.append((char)ch);

        is.close();
        closefc();
        isError = false;
      }
      catch (IOException ex)
      {
        try
        {
          is.close();
        }
        catch (Exception e) {}
        closefc();
        isError = true;
      }
    }
  }

// --------------------- Inner class WriteThread ------------------
  private class WriteThread extends Thread
  {

    private String data;
    private boolean append;

    WriteThread(String data, boolean append)
    {
      this.data = data;
      this.append = append;
    }

    public void run()
    {
      try
      {
        openfc();
        if (!fc.exists())
          fc.create();
        else
          if (!append)
            fc.truncate(0);
      }
      catch (IOException ex)
      {
        closefc();
        isError = true;
        return;
      }

      OutputStream os = null;

      try
      {
        if (append)
        {
          long size = fc.fileSize();
          os = fc.openOutputStream(size);
        }
        else
          os = fc.openOutputStream();
        os.write(data.getBytes());

        os.close();
        closefc();
        isError = false;
      }
      catch (IOException ex)
      {
        try
        {
          os.close();
        }
        catch (Exception e) {}
        closefc();
        isError = true;
      }
    }
  }

// ------------------------------------------------------------------
  private String path;
  private String fileURL;
  private FileConnection fc = null;
  private DataOutputStream dos = null;
  private DataInputStream dis = null;
  private OutputStream os = null;
  private InputStream is = null;
  private Enumeration em = null;
  private Enumeration emr = null;
  private volatile boolean isError = false;

  /**
   * Create a MFile instance with empty path.
   */
  public MFile()
  {
    this("");
  }

  /**
   * Creates a MFile instance with given path.
   * path is fully qualified with / separator. Directory pathes must end with a /<br>
   * e.g.<br>
   * root1/Pictures/image1.jpg  : path to file<br>
   * c:/Games/  : path to directory<br>
   * No connection to the given file/directory is yet established.
   * @param path of file or directory
   */
  public MFile(String path)
  {
    this.path = path;
    fileURL = "file:///" + path;
  }

  /**
   * Sets fully qualified path of file.
   * (Also modifies URL path.)
   * No connection to the given file/directory is yet established.
   */
  public void setPath(String path)
  {
    this.path = path;
    this.fileURL = "file:///" + path;
  }

  /**
   * Gets fully qualified path of file.
   * @return path
   */
  public String getPath()
  {
    return path;
  }

  /**
   * Gets file name part of the path.
   * @return the file name part or empty string,if path is a directory.
   */
  public String getName()
  {
    int index = path.lastIndexOf('/');
    if (index == path.length() - 1)  // at end
      return "";
    return path.substring(index + 1);
  }

  /**
   * Gets fully qualified URL path of file.
   * The URL path is composed as "file:///" + path.
   * @return URL path
   */
  public String getURL()
  {
    return fileURL;
  }

  /**
   * Checks if FileConnection (JSR 75) is supported.
   * @return true, if supported, otherwise false
   */
  public static boolean isFileConnectionSupported()
  {
    String version =
      System.getProperty("microedition.io.file.FileConnection.version");
    if (version == null)
      return false;
    return true;
  }

  /**
   * Opens a DataOutputStream used to write primitive data types to a file.
   * Uses the write<type>-Methods to write to the file.
   * The file connection and the file stream remain open,
   * until releaseConnection() (or another file operation method) is called.
   * If the stream is already open, nothing happens.
   * @return a reference to the opened stream or null, if operation fails.
   * @see #releaseConnection
   */
  public DataOutputStream getDataOutputStream()
  {
    try
    {
      if (fc == null)
        fc = (FileConnection)Connector.open(fileURL);
      if (dos == null)
        dos = fc.openDataOutputStream();
      return dos;
    }
    catch (IOException ex)
    {
      closefc();
      dos = null;
      return null;
    }
  }

  /**
   * Opens an OutputStream used to write data to a file.
   * The file connection and the file stream remain open,
   * until releaseConnection() (or another file operation method) is called.
   * If the stream is already open, nothing happens.
   * @return a reference to the opened stream or null, if operation fails.
   * @see #releaseConnection
   */
  public OutputStream getOutputStream()
  {
    try
    {
      if (fc == null)
        fc = (FileConnection)Connector.open(fileURL);
      if (os == null)
        os = fc.openOutputStream();
      return os;
    }
    catch (IOException ex)
    {
      closefc();
      os = null;
      return null;
    }
  }

  /**
   * Opens a DataInputStream used to read primitive data types from a file.
   * Uses the read<type>-Methods to read from the file and catch
   * the EOFException to stop reading.
   * The file connection and the file stream remain open,
   * until releaseConnection() (or another file operation method) is called.
   * If the stream is already open, nothing happens.
   * @return a reference to the opened stream or null, if operation fails.
   * @see #releaseConnection
   */
  public DataInputStream getDataInputStream()
  {
    try
    {
      if (fc == null)
        fc = (FileConnection)Connector.open(fileURL, Connector.READ);
      if (dis == null)
        dis = fc.openDataInputStream();
      return dis;
    }
    catch (IOException ex)
    {
      closefc();
      dis = null;
      return null;
    }
  }

  /**
   * Opens a InputStream used to read data from a file.
   * The file connection and the file stream remain open,
   * until releaseConnection() (or another file operation method) is called.
   * If the stream is already open, nothing happens.
   * @return a reference to the opened stream or null, if operation fails.
   * @see #releaseConnection
   */
  public InputStream getInputStream()
  {
    try
    {
      if (fc == null)
        fc = (FileConnection)Connector.open(fileURL, Connector.READ);
      if (is == null)
        is = fc.openInputStream();
      return is;
    }
    catch (IOException ex)
    {
      closefc();
      is = null;
      return null;
    }
  }

  /**
   * Writes string data to the text file.
   * If the file does not exist, it will be created.
   * If the file exists, all data is cleared.
   * Runs in a separate thread to improve performance.
   * All resources are released when finishing.
   * @return true, if successful, false if operation fails
   */
  public boolean write(String data)
  {
    return write(data, false);
  }

  /**
   * Appends string data to the end of the text file.
   * If the file does not exist, it will be created.
   * Runs in a separate thread to improve performance,
   * but waits for the thread to terminate.
   * All resources are released when finishing.
   * @return true, if successful, false if operation fails
   */
  public boolean append(String data)
  {
    return write(data, true);
  }

  private boolean write(String data, boolean append)
  {
    WriteThread wt = new WriteThread(data, append);
    wt.start();
    try
    {
      wt.join();
    }
    catch (InterruptedException ex) {}
    if (isError)
      return false;
    return true;
  }

  /**
   * Reads content of a text file.
   * Runs in a separate thread to improve performance,
   * but waits for the thread to terminate.
   * All resources are released when finishing.
   * @return content as string, empty if file has zero length, null if operation fails
   */
  public String read()
  {
    StringBuffer sb = new StringBuffer();
    ReadThread rt = new ReadThread(sb);
    rt.start();
    try
    {
      rt.join();
    }
    catch (InterruptedException ex) {}
    if (isError)
      return null;
    return sb.toString();
  }

  /**
   * Starts enumerating the files in the given directory.
   * Follow this procedure:<br>
   * 1. Set the path to the directory where files and subdirectories
   *    will be enumerated.<br>
   * 2. Call findFirst();  if it returns true, the  path is set
   *    to the first file/subdirectory.  Otherwise there is no file/subdirectory
   *    and you can't proceed.<br>
   * 3. Keep calling findNext() until it returns false,
   *    meaning there are no more files/directories.
   *    After every successful findNext() call, the path is set
   *    to the next file/subdirectory.<br>
   * Be aware of the side effect, that the path is changed,
   * if any file or subdirectory is found.<br>
   * @return true, if the given directory contains at least one file or subdirectory,
   * false otherwise (the path is unchanged).
   * @see #findNext
   */
  public boolean findFirst()
  {
    try
    {
      openfc();
      em = fc.list();
      closefc();
      if (em.hasMoreElements())
        setPath((String)em.nextElement());
      else
        return false;
    }
    catch (IOException ex)
    {
      em = null;
      closefc();
      return false;
    }
    return true;
  }

  /**
   * Continues enumerating the files after calling findFirst()
   * Be aware of the side effect, that the path is changed, if any file or
   * subdirectory is found.<br>
   * @see #findFirst
   * @return true, if the search for a next file or subdirectory is successful,
   * false, if there is no more file/subdirectory (the path is unchanged) or no preceding findFirst() was called.
   */
  public boolean findNext()
  {
    if (em == null)
      return false;
    if (em.hasMoreElements())
      setPath((String)em.nextElement());
    else
    {
      em = null;
      closefc();
      return false;
    }
    return true;
  }

  /**
   * Starts enumerating the file system roots.
   * Follow this procedure:<br>
   * 1. Set the path to "" or any other value (not used)<br>
   * 2. Call findFirstRoot(); if it returns true, the  path is set
   *    to the first detected root. Otherwise no root is found
   *    and you can't proceed.<br>
   * 3. Keep calling findNextRoot() until it returns false,
   *    meaning there are no more roots.
   *    After every successful findNext() call, the path is set
   *    to the next root.<br>
   * Be aware of the side effect, that the path is changed, if any root is found.
   * @return true, if the system reports at least one root,
   * false otherwise (the path is unchanged).
   * @see #findNextRoot
   */
  public boolean findFirstRoot()
  {
    emr = FileSystemRegistry.listRoots();
    if (emr.hasMoreElements())
      setPath((String)emr.nextElement());
    else
    {
      emr = null;
      return false;
    }
    return true;
  }

  /**
   * Continues enumerating the roots after calling findFirstRoot()
   * Be aware of the side effect, that the path is changed, if a root is found.
   * @see #findFirstRoot
   * @return true, if the search for a next root is successful,
   * false, if there are no more roots (the path is unchanged) or no preceding findFirstRoot() was called.
   */
  public boolean findNextRoot()
  {
    if (emr == null)
      return false;
    if (emr.hasMoreElements())
      setPath((String)emr.nextElement());
    else
    {
      emr = null;
      return false;
    }
    return true;
  }

  /**
   * Checks if the path corresponds to an existing file or directory.
   * @return true, if the file/directory exists, otherwise false
   */
  public boolean exists()
  {
    try
    {
      openfc();
    }
    catch (IOException ex)
    {
      return false;
    }
    boolean rc = fc.exists();
    closefc();
    return rc;
  }

  /**
   * Checks if the path corresponds to an existing file (not a directory).
   * @return true, if the file exists, otherwise false
   */
  public boolean isFile()
  {
    return exists() && (!isDirectory());
  }

  /**
   * Checks if the path corresponds to an existing directory.
   * @return true, if the directory exists, otherwise false
   */
  public boolean isDirectory()
  {
    try
    {
      openfc();
    }
    catch (IOException ex)
    {
      return false;
    }
    boolean rc = fc.isDirectory();
    closefc();
    return rc;
  }

  /**
   * Removes file or directory.
   * To remove a directory, it must be empty.
   * @return true, if successful, false, if file or directory does not exist or operation fails
   */
  public boolean remove()
  {
    try
    {
      openfc();
      if (fc != null)
        fc.delete();
    }
    catch (IOException ex)
    {
      closefc();
      return false;
    }
    closefc();
    return true;
  }

  /**
   * Renames file or directory.
   * The given name must be a valid file or directory name without any further path information.
   * The MFile's path is not modified.
   * @return true, if successful, false, if operation fails
   */
  public boolean rename(String newName)
  {
    try
    {
      openfc();
      if (fc != null)
        fc.rename(newName);
    }
    catch (IOException ex)
    {
      closefc();
      return false;
    }
    closefc();
    return true;
  }

  /**
   * Clears all data.
   * After clearing the filesize is 0.
   * @return true, if successful, false, if operation fails
   */
  public boolean clear()
  {
    try
    {
      openfc();
      if (fc != null)
        fc.truncate(0);
    }
    catch (IOException ex)
    {
      closefc();
      return false;
    }
    closefc();
    return true;
  }

  /**
   * Creates a directory corresponding to the MFile's path.
   * Only one directory level is allowed. A directory path must end with /.
   * @return true, if successful, false, if operation fails
   */
  public boolean createDir()
  {
    try
    {
      openfc();
      if (fc != null)
        fc.mkdir();
    }
    catch (IOException ex)
    {
      closefc();
      return false;
    }
    closefc();
    return true;
  }

  /**
   * Creates a file with zero length.
   * All directories in the MFile's path must exist, before the file may be created.
   * If the file already exists, it is deleted first.
   * @return true, if successful, otherwise false
   */
  public boolean createFile()
  {
    try
    {
      openfc();
      if (fc != null)
      {
        if (fc.exists())
          fc.truncate(0);
        else
          fc.create();
      }
    }
    catch (IOException ex)
    {
      closefc();
      return false;
    }
    closefc();
    return true;
  }

  /**
   * Checks if the file/directory is readable.
   * @return true, if readable, otherwise false.
   */
  public boolean canRead()
  {
    try
    {
      openfc();
    }
    catch (IOException ex)
    {
      return false;
    }
    boolean rc = fc.canRead();
    closefc();
    return rc;
  }

  /**
   * Checks if the file/directory is writable.
   * @return true, if writable, otherwise false.
   */
  public boolean canWrite()
  {
    try
    {
      openfc();
    }
    catch (IOException ex)
    {
      return false;
    }
    boolean rc = fc.canWrite();
    closefc();
    return rc;
  }

  /**
   * Checks if the file/directory is hidden.
   * @return true, if hidden, otherwise false.
   */
  public boolean isHidden()
  {
    try
    {
      openfc();
    }
    catch (IOException ex)
    {
      return false;
    }
    boolean rc = fc.isHidden();
    closefc();
    return rc;
  }

  /**
   * Sets the hidden attribute of file or directory to the given value.
   * @return true, if successful, otherwise false.
   */
  public boolean setHidden(boolean hidden)
  {
    try
    {
      openfc();
      if (fc != null)
        fc.setHidden(hidden);
    }
    catch (IOException ex)
    {
      closefc();
      return false;
    }
    closefc();
    return true;
  }

  /**
   * Sets the read attribute of file or directory to the given value.
   * @return true, if successful, otherwise false.
   */
  public boolean setReadable(boolean readable)
  {
    try
    {
      openfc();
      if (fc != null)
        fc.setReadable(readable);
    }
    catch (IOException ex)
    {
      closefc();
      return false;
    }
    closefc();
    return true;
  }

  /**
   * Sets the write attribute of file or directory to the given value.
   * @return true, if successful, otherwise false.
   */
  public boolean setWritable(boolean writable)
  {
    try
    {
      openfc();
      if (fc != null)
        fc.setWritable(writable);
    }
    catch (IOException ex)
    {
      closefc();
      return false;
    }
    closefc();
    return true;
  }

  /**
   * Gets the size of the file.
   * @return File size in bytes or -1 if operation fails
   */
  public long getFileSize()
  {
    long rc = -1;
    try
    {
      openfc();
      if (fc != null)
        rc = fc.fileSize();
    }
    catch (IOException ex)
    {
      closefc();
      return -1;
    }
    closefc();
    return rc;
  }

  /**
   * Determines the free space on the file system, where the MFile's path resides on.
   * @return the free space in bytes or -1 when operation fails
   */
  public long getAvailableSpace()
  {
    try
    {
      openfc();
    }
    catch (IOException ex)
    {
      return -1;
    }
    long rc = fc.availableSize();
    closefc();
    return rc;
  }

  /**
   * Determines the amount of space used by all files on the file system, where the MFile's path resides on.
   * If includeSubDirs is true, include all subdirectories.
   * The MFile's path must point to an existing directory (terminated with /).
   * @return the amount of bytes used by the file system, -1 of operation fails.
   */
  public long getUsedSpace(boolean includeSubDirs)
  {
    long rc = -1;
    try
    {
      openfc();
      if (fc != null)
        rc = fc.directorySize(includeSubDirs);
    }
    catch (IOException ex)
    {
      closefc();
      return -1;
    }
    closefc();
    return rc;
  }

  /**
   * Closes an eventually opened connection and open a new one.
   * Calls releaseConnection() to close the connection.
   * @return Reference to the FileConnection or null, if operation fails
   * @see #releaseConnection
   */
  public FileConnection openConnection()
  {
    try
    {
      openfc();
      return fc;
    }
    catch (IOException ex)
    {
      return null;
    }
  }

  /**
   * Releases all allocated resources (FileConnection, DataInputStream, DataOutputStream).
   * If resources are not actually in use, nothing happens.
   */
  public void releaseConnection()
  {
    try { fc.close();} catch (Exception e) {}
    try {is.close();} catch (Exception e) {}
    try {os.close();} catch (Exception e) {}
    try {dis.close();} catch (Exception e) {}
    try {dos.close();} catch (Exception e) {}
    fc = null;
    is = null;
    os = null;
    dis = null;
    dos = null;
  }

  private void openfc() throws IOException
  {
    if (fc != null)
      fc.close();
    fc = (FileConnection)Connector.open(fileURL);
  }

  private void closefc()
  {
    if (fc != null)
    {
      try {fc.close();} catch (IOException e) {}
      fc = null;
    }
  }
}
