// MList.java

/*
  This software is part of Gidlet framework.
  It is Open Source Free Software, so you may
    - run the code for any purpose
    - study how the code works and adapt it to your needs
    - integrate all or parts of the code in your own programs
    - redistribute copies of the code
    - improve the code and release your improvements to the public
  However the use of the code is entirely your responsibility.
*/

package ch.aplu.gidlet;

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import java.util.Hashtable;
import java.util.Timer;
import java.util.TimerTask;
import ch.aplu.gidlet.*;

/**
 * List display with simple input and output items.
 *
 * A soft button 'Exit' is shown. Hitting the corresponding key will
 * call the Gidlet's doExit(). If you do not need this button,
 * invoke removeExitButton().<br>
 * If you implement your own CommandListener and call setCommandListener(), the internal
 * CommandListener is disabled.<br>
 *
 * main() serves as entry point for the user programm. It is called in the run-method
 * of a separate thread and should terminate to avoid non-terminated threads.<br>
 *
 * Creation may take place within main() or while declaring instance variable of
 * the application program.<br>
 *
 * Example:<p>
 * <code><pre>
 *
 * import ch.aplu.gidlet.*;
 * import javax.microedition.lcdui.*;
 *
 * public class ListSelectGidlet extends Gidlet
 * {
 *   private MList list = new MList("Email");
 *
 *   public void main()
 *   {
 *     list.append("Read", null);
 *     list.append("Reply", null);
 *     list.append("Delete", null);
 *     list.show();
 *   }
 *
 *   public void doOk()
 *   {
 *     int index = list.getSelectedIndex();
 *     String selection = list.getString(index);
 *     showMessage(selection);
 *   }
 * }
 *</pre></code>
 */
public class MList extends javax.microedition.lcdui.List
{
// ------------------ Inner class MyCommandListener -------------
private class MyCommandListener implements CommandListener
{
  public void commandAction(Command c, Displayable d)
  {
    if (c == exitCommand)
      gidlet.doExit();
    if (okCommand != null && c == okCommand)
    {
      gidlet.doOk();
      gidlet.doOk(d);
      if (gidlet.isSleeping())
        gidlet.wakeUp();
    }
  }
}


// ------------------ Inner class TimerTask ---------------------
private class MyTimerTask extends TimerTask
{
  public void run()
  {
    waitVisible = false;
  }
}
// ------------------ End of inner classes ----------------------

  private Gidlet gidlet;
  private Command exitCommand = new Command("Exit", Command.EXIT, 1);
  private Command okCommand = null;
  private boolean waitVisible;

  /**
   * Constructs a MList with given title and list type IMPLICIT.
   * The display is initially invisible and should be made visible
   * by calling show().
   * If an EXCLUSIVE (radio buttons) or MULTIPLE list (check boxes)
   * is needed, use a ChoiceGroup item instead.
   */
  public MList(String title)
  {
    super((title == null) ? "" : title, List.IMPLICIT);
    this.gidlet = Gidlet.myGidlet;
    addCommand(exitCommand);
    setCommandListener(new MyCommandListener());
  }

  /**
   * Constructs a MList with no title and list type IMPLICIT.
   * The display is initially invisible and should be made visible
   * by calling show().
   * If an EXCLUSIVE (radio buttons) or MULTIPLE list (check boxes)
   * is needed, use a ChoiceGroup item instead.
   */
  public MList()
  {
    this("");
  }

  /**
   * Shows the form display if it is not yet visible or
   * it was previously hidden by another display.
   * Because it is not garanteed that the diplay is actually visible when
   * Display.setCurrent() returns, wait for a maximum of 5 s for the display
   * to get shown.
   *
   */
  public void show()
  {
    show(true);
  }

  /**
   * Shows the list display if it is not yet visible or
   * it was previously hidden by another display.
   * If wait = true, block until the display becomes actually visible, up
   * to a maximum auf 5 seconds. This may be useful to observe the start of
   * an animation.
   */
  public void show(boolean wait)
  {
    gidlet.display.setCurrent(this);
    if (wait)
    {
      waitVisible = true;
      Timer t = new Timer();
      t.schedule(new MyTimerTask(), 5000);
      while (waitVisible && !isShown())
        Gidlet.delay(100);
      t.cancel();
    }
  }

  /**
   * Adds an OK soft button with label "OK". When pressed, the
   * Gidlet's callback notification doOk() is called unless you
   * register your own CommandListener.
   */
  public void addOkButton()
  {
    addOkButton("OK");
  }

  /**
   * Adds an OK soft button with given label. When pressed, the
   * Gidlet's callback notification doOk() is called unless you
   * register your own CommandListener.
   */
  public void addOkButton(String label)
  {
    okCommand = new Command(label, Command.OK, 1);
    addCommand(okCommand);
  }

  /**
   * Removes Exit soft button.
   * Gidlet's callback notification doExit() is disabled.
   */
  public void removeExitButton()
  {
   removeCommand(exitCommand);
  }

  /**
   * Removes Ok soft button.
   * Gidlet's callback notification doOk() is disabled.
   */
  public void removeOkButton()
  {
   removeCommand(okCommand);
  }
}
