// Turtle.java

/*
  This software is part of the Gidlet framework.
  It is Open Source Free Software, so you may
    - run the code for any purpose
    - study how the code works and adapt it to your needs
    - integrate all or parts of the code in your own programs
    - redistribute copies of the code
    - improve the code and release your improvements to the public
  However the use of the code is entirely your responsibility.
*/

package ch.aplu.gidlet;

import javax.microedition.lcdui.*;
import ch.aplu.gidlet.*;

/**
 * Logo-like turtle that moves on a graphics display, called playground.
 * The coordinate system has it's reference point in the middle of the screen
 * with x-coordinates to the right and y-coordinates upward both in the range
 * -200..200 (doubles). For simplicity the turtle shape is a triangle. Each
 * turtle instance has it's own playground. The most recent turtle's playground
 * is visible, while other playgrounds or Gui windows become hidden. The hidden
 * windows may be updated in any way, but the change is only visible, when
 * calling their show() method. <br>
 * Default colors:<br>
 * background (playground) color: black<br>
 * turtle color: light gray<br>
 * turtle pen color: white<br>
 * fill color: white<br><br>
 *
 *
 * Example:<p>
 * <code><pre>
 *import ch.aplu.gidlet.*;
 *
 *public class TurtleGidlet extends Gidlet
 *{
 *  public void main()
 *  {
 *    Turtle t = new Turtle();
 *
 *    for (int i = 0; i < 18; i++)
 *    {
 *      for (int k = 0; k < 4; k++)
 *        t.forward(100).left(90);
 *      t.left(20);
 *    }
 *  }
 *}
 * </pre></code>
 */
public class Turtle
{
  // ---------------------- Inner class PointD ------------------
  protected class PointD
  {
    public double x;
    public double y;
    public int border;

    public PointD(double x, double y, int border)
    {
      this.x = x;
      this.y = y;
      this.border = border;
    }
  }

  // ---------------------- Inner interface FillMode ------------
  private interface FillMode
  {
    int fillOff  = 0;
    int fillVert  = 1;
    int fillHorz  = 2;
    int fillPoint = 3;
  }

  // ---------------------- Inner interface Border --------------
  private interface Border
  {
    int top  = 0;
    int bottom  = 1;
    int left  = 2;
    int right = 3;
  }
  // ---------------------- End of inner classes/interfaces -----

  // Properties of turtle
  private double x, y; // Current graphics cursor position
  private double _x, _y; // Current graphics cursor position wrapped to range
  private double phi; // Current angle clockwise from north in degrees
  private int turtleColor; // Current color
  private int penColor; // Current color
  private int fillColor; // Color for filling
  private int bgColor;  // Background color
  private boolean isVisible; // Turtle shown/hidden
  private boolean isTraceShown; // Pen up/down
  private boolean isWrap; // Wrap/unwrap at borders to opposite side
  private int fillMode; // fillOff, fillHorz, fillVert, fillPoint
  private int delayTime; // Delay turtle movements
  private double fillAnchorX, fillAnchorY;  // Anchor of filling rubber band
  private double fillLine;  // Coordinates of filling lines
  private String title; // Playground's title

  private Turtle parent;
  private Playground pg;
  protected TurtleShape shape;  // Playground must access

  private final static int defaultBgColor = Gidlet.WHITE;
  private final static int defaultTurtleColor = Gidlet.CYAN;
  private final static int defaultPenColor = Gidlet.BLUE;
  private final static int defaultFillColor = Gidlet.BLUE;
  private final static String defaultTitle = "Turtle Playground";
  private final static int defaultTitleColor = Gidlet.BLACK;
  private final static double range = 200;

  /**
   * Creates a turtle clone that has exactly the same properties and
   * traces as the given turtle, but the playground has the given title.
   * Its playground is initially hidden and becomes the current  window
   * by calling showPlayground().
   * @see #clone
   * @see #showPlayground
   */
  public Turtle(Turtle t, String title)
  {
    this(title, t.bgColor, t.turtleColor, t.isVisible, false);
    parent = t;
    x = t.x;
    y = t.y;
    phi = t.phi;
    _x = t._x;
    _y = t._y;
    fillAnchorX = t.fillAnchorX;
    fillAnchorY = t.fillAnchorY;
    fillLine = t.fillLine;
    penColor = t.penColor;
    fillColor = t.fillColor;
    isTraceShown = t.isTraceShown;
    isWrap = t.isWrap;
    fillMode = t.fillMode;
    delayTime = t.delayTime;

    if (Thread.currentThread().getName().equals("Gidlet-Thread"))
      init();
    else // defer initialization
      (Gidlet.myGidlet.turtles).addElement(this);
   }

  /**
   * Initializes Turtle. Called by main() if Turtle is created before main() is called.
   */
  protected void init()
  {
    // Copy parent's trace buffer to child's trace buffer
    pg.getTraceBuffer().drawImage(parent.pg.getTraceImage(), 0, 0, 0);

    // Copy child's trace buffer to child's turtle buffer and add turtle image
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);
  }

 /**
  * Creates a turtle clone that has exactly the same properties, playground title
  * and traces as the given turtle. Its playground is initially hidden and
  * becomes the current  window by calling showPlayground().
  * @see #clone
  * @see #showPlayground
  */
  public Turtle(Turtle t)
  {
    this(t, t.title);
  }

  /**
   * Constructs a Turtle instance with given playground title,
   * background (playground) color, turtle color
   * and visibility.<br>
   * If showPlayground = false, the turtle's window is initially invisible
   * and must made visible by calling showPlayground().
   * Turtle is at position (0,0) looking north. Wrapping is off
   * and pen is down.
   */
  public Turtle(String title, int bgColor, int color, boolean visible, boolean showPlayground)
  {
    // Initialize properties
    x = 0;
    y = 0;
    phi = 0;
    _x = 0;
    _y = 0;
    fillAnchorX = 0;
    fillAnchorY = 0;
    fillLine = 0;
    turtleColor = color;
    penColor = defaultPenColor;
    fillColor = defaultFillColor;
    this.bgColor = bgColor;
    isVisible = visible;
    isTraceShown = true;
    isWrap = false;
    fillMode = FillMode.fillOff;
    delayTime = 0;
    this.title = title;

    pg = new Playground(this, title, range, bgColor, color, penColor, defaultTitleColor,
                        visible, showPlayground);
  }

  /**
   * Constructs a Turtle instance with given playground title,
   * background (playground) color, turtle color
   * and visibility.<br>
   * Turtle is at position (0,0) looking north. Wrapping is off
   * and pen is down.
   */
  public Turtle(String title, int bgColor, int color, boolean visible)
  {
    this(title, bgColor, color, visible, true);
  }

  /**
   * Constructs a Turtle instance with given background (playground) color, turtle color
   * and visibility.<br>
   * Turtle is at position (0,0) looking north. Wrapping is off
   * and pen is down.
   */
  public Turtle(int bgColor, int color, boolean visible)
  {
    this(defaultTitle, bgColor, color, visible, true);
  }

  /**
   * Constructs a Turtle instance with default background (playground) color,
   * given turtle color and visibility.<br>
   * Turtle is at position (0,0) looking north. Wrapping is off
   * and pen is down.
   */
  public Turtle(int color, boolean visible)
  {
    this(defaultTitle, defaultBgColor, color, visible, true);
  }

  /**
   * Constructs a Turtle instance with default background (playground) color and
   * given turtle color.<br>
   * Turtle is visible at position (0,0) looking north. Wrapping is off
   * and pen is down.
   */
  public Turtle(int color)
  {
    this(defaultTitle, defaultBgColor, color, true, true);
  }

  /**
   * Constructs a Turtle instance with given visibility, default background (playground) color
   * and default turtle color.<br>
   * Turtle is at position (0,0) looking north. Wrapping is off
   * and pen is down.
   */
  public Turtle(boolean visible)
  {
    this(defaultTitle, defaultBgColor, defaultTurtleColor, visible, true);
  }

  /**
   * Constructs a Turtle instance with given playground title, default background (playground) color and
   * default turtle color.<br>
   * Turtle is visible at position (0,0) looking north. Wrapping is off
   * and pen is down.
   */
  public Turtle(String title)
  {
    this(title, defaultBgColor, defaultTurtleColor, true, true);
  }

  /**
   * Constructs a Turtle instance with default background (playground) color and
   * default turtle color.<br>
   * Turtle is visible at position (0,0) looking north. Wrapping is off
   * and pen is down.
   */
  public Turtle()
  {
    this(defaultTitle, defaultBgColor, defaultTurtleColor, true, true);
  }

  /**
   * Moves turtle the given distance forward.
   * Returns the Turtle reference.
   */
  public Turtle forward(double s)
  {
    double p = phi / 180 * Math.PI;
    double deltax = s * Math.sin(p);
    double deltay = s * Math.cos(p);
    double xnew = x + deltax;
    double ynew = y + deltay;
    double _xnew = _x + deltax;  // Maybe outside range
    double _ynew = _y + deltay;  // ditto

    put(xnew, ynew, _xnew, _ynew);
    Gidlet.delay(delayTime);
    return this;
  }

  private void put(double xnew, double ynew, double _xnew, double _ynew)
  {
    if (isTraceShown)
    {
      if (isWrap)
      {
        switch (fillMode)
        {
          case FillMode.fillOff:
            pg.move(_x, _y);
            pg.draw(_xnew, _ynew);  // draw even if outside range
            break;

          case FillMode.fillPoint:
            pg.fillTriangle(fillAnchorX, fillAnchorY,
                            _x, _y,
                            _xnew, _ynew, fillColor);
            break;

          case FillMode.fillHorz:
            pg.fillRectangle(_x, fillLine, _xnew, Math.min(_y, _ynew), fillColor);
            pg.fillTriangle(_x, _y,
                            _xnew, _ynew,
                            _ynew > _y ? _xnew : _x,
                            _ynew > _y ? _y : _ynew,
                            fillColor);
           break;

          case FillMode.fillVert:
            pg.fillRectangle(fillLine, _y, Math.min(_x, _xnew), _ynew, fillColor);
            pg.fillTriangle(_x, _y,
                            _xnew, _ynew,
                            _xnew > _x ? _x : _xnew,
                            _xnew > _x ? _ynew : _y,
                            fillColor);
           break;
        }

        PointD pt = getIntersection(_x, _y, _xnew, _ynew);
        if (pt != null)
        {
          switch (pt.border)
          {
            case Border.right:
              _x = _x - 2*range;
              break;

            case Border.left:
              _x = _x + 2*range;
              break;

            case Border.top:
              _y = _y - 2*range;
              break;

            case Border.bottom:
              _y = _y + 2*range;
              break;
          }

          switch (fillMode)
          {
            case FillMode.fillOff:
              pg.move(_x,  _y);
              pg.draw(shape.wrapToRange(_xnew), shape.wrapToRange(_ynew));
              break;

            case FillMode.fillPoint:
              pg.fillTriangle(fillAnchorX, fillAnchorY,
                              _x, _y,
                              shape.wrapToRange(_xnew), shape.wrapToRange(_ynew),
                              fillColor);
              break;

            case FillMode.fillHorz:
              double _xwrap = shape.wrapToRange(_xnew);
              double _ywrap = shape.wrapToRange(_ynew);
              pg.fillRectangle(_x, fillLine, _xwrap, Math.min(_y, _ywrap), fillColor);
              pg.fillTriangle(_x, _y,
                              _xwrap, _ywrap,
                              _ywrap > _y ? _xwrap : _x,
                              _ywrap > _y ? _y : _ywrap,
                              fillColor);
             break;

           case FillMode.fillVert:
              _xwrap = shape.wrapToRange(_xnew);
              _ywrap = shape.wrapToRange(_ynew);
              pg.fillRectangle(fillLine, _y, Math.min(_x, _xwrap), _ywrap, fillColor);
              pg.fillTriangle(_x, _y,
                              _xwrap, _ywrap,
                              _xwrap > _x ? _x : _xwrap,
                              _xwrap > _x ? _ywrap : _y,
                              fillColor);
             break;
          }
        }
      }
      else  // no wrap
      {
        switch (fillMode)
        {
          case FillMode.fillOff:
            pg.move(x, y);
            pg.draw(xnew, ynew);
            break;

          case FillMode.fillPoint:
            pg.fillTriangle(fillAnchorX, fillAnchorY,
                            x, y,
                            xnew, ynew,
                            fillColor);
            break;

          case FillMode.fillHorz:
            pg.fillRectangle(x, fillLine, xnew, Math.min(y, ynew), fillColor);
            pg.fillTriangle(x, y,
                            xnew, ynew,
                            ynew > y ? xnew : x,
                            ynew > y ? y : ynew,
                            fillColor);
            break;

          case FillMode.fillVert:
            pg.fillRectangle(fillLine, y, Math.min(x, xnew), ynew, fillColor);
            pg.fillTriangle(x, y,
                            xnew, ynew,
                            xnew > x ? x : xnew,
                            xnew > x ? ynew : y,
                            fillColor);
            break;
        }
      }
    }

    x = xnew;
    y = ynew;
    _x = shape.wrapToRange(_xnew);
    _y = shape.wrapToRange(_ynew);
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);
  }

  /**
   * Short for forward().
   * @see #forward
   */
  public Turtle fd(double s)
  {
    return forward(s);
  }

  /**
   * Moves turtle the given distance backward.
   * Returns the Turtle reference.
   */
  public Turtle back(double s)
  {
    return forward(-s);
  }

  /**
   * Short for back().
   * @see #back
   */
  public Turtle bk(double s)
  {
    return back(s);
  }

  /**
   * Turns turtle the given angle (in degrees) to the right.
   * Returns the Turtle reference.
   */
  public Turtle right(double angle)
  {
    phi = phi + angle;
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);
    Gidlet.delay(delayTime);
    return this;
  }

  /**
   * Short for right().
   * @see #right
   */
  public Turtle rt(double angle)
  {
    return right(angle);
  }

  /**
   * Turns turtle the given angle (in degrees) to the left.
   * Returns the Turtle reference.
   */
  public Turtle left(double angle)
  {
    phi = phi - angle;
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);
    Gidlet.delay(delayTime);
    return this;
  }

  /**
   * Short for left().
   * @see #left
   */
  public Turtle lt(double angle)
  {
    return left(angle);
  }

  /**
   * Returns the x-coordinate. If wrapping is on, return the wrapped coordinate within
   * the playground range.
   */
  public double getX()
  {
    return _x;
  }

  /**
   * Returns the y-coordinate. If wrapping is on, return the wrapped coordinate within
   * the playground range.
   */
  public double getY()
  {
    return _y;
  }
  /**
   * Returns the x-coordinate. Even if wrapping is on, return the unwrapped coordinate not
   * bound to the playground range.
   */

  public double _getX()
  {
    return x;
  }

  /**
   * Returns the y-coordinate. If wrapping is on, return the unwrapped coordinate not
   * bound the playground range.
   */
  public double _getY()
  {
    return y;
  }

  /**
   * Returns the turtle color.
   */
  public int getColor()
  {
    return turtleColor;
  }

  /**
   * Emits a beep.
   * Returns the Turtle reference.
   */
  public Turtle beep()
  {
    Gidlet.beep(0);
    return this;
  }

  /**
   * Erases traces and text, but let turtle where it is. Stamped turtles are erased.
   * If the turtle is hidden, it remains hidden.
   * Returns the Turtle reference.
   */
  public Turtle clean()
  {
    shape.stampedTurtles.removeAllElements();
    pg.clear(false);
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);
    return this;
  }

  /**
   * Clears the entire playground. Reset all turtle properties to starting values.
   * Returns the Turtle reference.
   */
  public Turtle clear()
  {
    x = 0;
    y = 0;
    phi = 0;
    _x = 0;
    _y = 0;
    fillAnchorX = 0;
    fillAnchorY = 0;
    fillLine = 0;
    penColor = defaultPenColor;
    fillColor = defaultFillColor;
    isTraceShown = true;
    isWrap = false;
    fillMode = FillMode.fillOff;
    delayTime = 0;

    shape.stampedTurtles.removeAllElements();
    pg.clear(true);
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);
    return this;
  }

  /**
   * Returns the distance from the current position to the given position.
   */
  public double distance(double x, double y)
  {
    return Math.sqrt((x - this.x) * (x - this.x) + (y - this.y) * (y - this.y));
  }

  /**
   * Returns the angle of vision (angle in respect to north, positive clockwise).
   */
  public double heading()
  {
    return phi;
  }

 /**
   * Makes turtle invisible.
   * Returns the Turtle reference.
   */
  public Turtle hideTurtle()
  {
    isVisible = false;
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);

    return this;
  }

  /**
   * Short for hideTurtle().
   * @see #hideTurtle
   */
  public Turtle ht()
  {
    return hideTurtle();
  }

  /**
   * Returns true if turtle is invisible (hidden), otherwise false.
   */
  public boolean isHidden()
  {
    return !isVisible;
  }

  /**
   * Returns true if turtle's pen is up, otherwise false.
   */
  public boolean isPenUp()
  {
    return !isTraceShown;
  }

  /**
   * Lower pen, so that the trace is drawn.
   * Returns the Turtle reference.
   */
  public Turtle penDown()
  {
    isTraceShown = true;
    return this;
  }

  /**
   * Short for penDown().
   * @see #penDown
   */
  public Turtle pd()
  {
    return penDown();
  }

  /**
   * Lifts pen, so that the trace is not drawn.
   * Returns the Turtle reference.
   */
  public Turtle penUp()
  {
    isTraceShown = false;
    return this;
  }

  /**
   * Short for penUp().
   * @see #penUp
   */
  public Turtle pu()
  {
    return penUp();
  }

  /**
   * Sets turtle's color to the given color.
   * Returns the Turtle reference.
   */
  public Turtle setColor(int color)
  {
    turtleColor = color;
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);

    return this;
  }

  /**
   * Sets turtle's color to the given color in RGB-format.
   * Returns the Turtle reference.
   */
  public Turtle setColor(int r, int g, int b)
  {
    int rgb = (r << 16) + (g << 8) + b;
    return setColor(rgb);
  }

  /**
   * Sets turtle's pen color to the given color.
   * Returns the Turtle reference.
   */
  public Turtle setPenColor(int color)
  {
    penColor = color;
    pg.setTracePenColor(color);
    return this;
  }

  /**
   * Sets turtle's pen color to the given color in RGB-format.
   * Returns the Turtle reference.
   */
  public Turtle setPenColor(int r, int g, int b)
  {
    int rgb = (r << 16) + (g << 8) + b;
    return setPenColor(rgb);
  }

  /**
   * Changes pen color to the background color.
   * Returns the Turtle reference.
   */
  public Turtle penErase()
  {
    return setPenColor(bgColor);
  }

  /**
   * Short for penErase().
   * @see #penErase
   */
  public Turtle pe()
  {
    return penErase();
  }

  /**
   * Sets the angle of vision (angle in respect to north, positive clockwise) to
   * the given value (in degrees).
   * Returns the Turtle reference.
   */
  public Turtle setHeading(double angle)
  {
    phi = angle;
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);
    return this;
  }

  /**
   * Short for setHeading().
   * @see #setHeading
   */
  public Turtle setH(double angle)
  {
    return setHeading(angle);
  }

  /**
   * Sets turtle to the given coordinates; if pen is down, a trace from
   * the last to the given position is drawn.
   * The angle of vision is not changed.
   * Returns the Turtle reference.
   */
  public Turtle setXY(double x, double y)
  {
    put(x, y, shape.wrapToRange(x), shape.wrapToRange(y));
    Gidlet.delay(delayTime);
    return this;
  }

  /**
   * Sets turtle to the given x-position, while the y-coordinate remains the same;
   * if pen is down, a trace from the last to the given position is drawn.
   * The angle of vision is not changed.
   * Returns the Turtle reference.
   */
  public Turtle setX(double x)
  {
    return setXY(x, y);
  }

  /**
   * Sets turtle to the given y-position, while the x-coordinate remains the same;
   * if pen is down, a trace from the last to the given position is drawn.
   * The angle of vision is not changed.
   * Returns the Turtle reference.
   */
  public Turtle setY(double y)
  {
    return setXY(x, y);
  }

  /**
   * Sets turtle to the given coordinates, without drawing a trace.
   * The angle of vision is not changed.
   * Returns the Turtle reference.
   */
  public Turtle setPos(double x, double y)
  {
    this.x = x;
    this.y = y;
    this._x = shape.wrapToRange(x);
    this._y = shape.wrapToRange(y);
    return this;
  }

  /**
   * Sets turtle to the given x-position without drawing a trace
   * while the y-coordinate remains unchanged.
   * The angle of vision is not changed.
   * Returns the Turtle reference.
   */
  public Turtle setPosX(double x)
  {
    return setPos(x, y);
  }

  /**
   * Sets turtle to the given y-position, without drawing a trace
   * while the x-coordinate remains unchanged.
   * The angle of vision is not changed.
   * Returns the Turtle reference.
   */
  public Turtle setPosY(double y)
  {
    return setPos(x, y);
  }



  /**
   * Makes turtle visible.
   * Returns the Turtle reference.
   */
  public Turtle showTurtle()
  {
    isVisible = true;
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);
    return this;
  }

  /**
   * Short for showTurtle().
   * @see #showTurtle
   */
  public Turtle st()
  {
    return showTurtle();
  }

  /**
   * Converts from pixel x-coordinates to turtle x-coordinates.
   */
  public double toTurtleX(int x)
  {
    return pg.toWindowX(x);
  }

  /**
   * Converts from pixel y-coordinates to turtle y-coordinates.
   */
  public double toTurtleY(int y)
  {
    return pg.toWindowY(y);
  }

  /**
   * Converts from turtle x-coordinates to pixel x-coordinates.
   */
  public int toPixelX(double x)
  {
    return pg.toPixelX(x);
  }

  /**
   * Converts from turtle y-coordinates to pixel y-coordinates.
   */
  public int toPixelY(double y)
  {
    return pg.toPixelY(y);
  }

  /**
   * Returns version information.
   */
  public String version()
  {
    return SharedConstants.VERSION;
  }

  /**
   * Causes turtle to wrap around playground borders, e.g. when it leaves the
   * playground on one border, it reappears on the opposite border. The turlte
   * should not move further than to adjacent playgrounds in one single operation.
   * Returns the Turtle reference.
   */
  public Turtle wrap()
  {
    isWrap = true;
    return this;
  }

  /**
   * Returns true, if turtle is in wrap mode, otherwise false.
   * @see #wrap
   */
  public boolean isWrap()
  {
    return isWrap;
  }

  /**
   * Causes turtle not to wrap around playground borders, e.g. when it leaves the
   * playground it disappaers.
   * Returns the Turtle reference.
   * @see #wrap
   */
  public Turtle clip()
  {
    isWrap = false;
    return this;
  }

  /**
   * Draws the given text using the current pen color and text font at the current turtle position.
   * The given index is used to reference the text for erasing. If the label is already
   * in use, the previous text is overwritten.
   * Returns the Turtle reference.
   * @see #eraseLabel
   */
  public Turtle drawLabel(int index, String text)
  {
    pg.addLabel(text, pg.toPixelX(_x) + 2, pg.toPixelY(_y) - 2, index, penColor);
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);
    return this;
  }

  /**
   * Erases the text with given index. If text with given index does not
   * exist, nothing happens.
   * Returns the Turtle reference.
   * @see #drawLabel
   */
  public Turtle eraseLabel(int index)
  {
    pg.removeLabel(index);
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);
    return this;
  }

  /**
   * Sets the current text font for drawing labels. If the font is changed, also
   * existing labels are affected.
   * Returns the Turtle reference.
   * Default font is FACE_SYSTEM, STYLE_PLAIN, SIZE_MEDIUM)
   * @see #drawLabel
   * @param face one of Font.FACE_SYSTEM, Font.FACE_MONOSPACE, or Font.FACE_PROPORTIONAL<br>
   * @param style Font.STYLE_PLAIN, or a combination of Font.STYLE_BOLD, Font.STYLE_ITALIC, and Font.STYLE_UNDERLINED<br>
   * @param size one of Font.SIZE_SMALL, Font.SIZE_MEDIUM, or Font.SIZE_LARGE
   */
  public Turtle setFont(int face, int style, int size)
  {
    Font font = Font.getFont(face, style, size);
    pg.getTurtleBuffer().setFont(font);
    return this;
  }

  /**
   * Registers an KeyListener to get a notification when a button is clicked.
   * Returns the Turtle reference.
   */
  public Turtle addKeyListener(KeyListener keyListener)
  {
    pg.addKeyListener(keyListener);
    return this;
  }

  /**
   * Removes a registered KeyListener.
   * Returns the Turtle reference.
   */
  public Turtle removeKeyListener()
  {
    pg.removeKeyListener();
    return this;
  }

  /**
   * Sets the given time delay in ms after each turtle movement.
   * Returns the Turtle reference.
   */
  public Turtle setDelay(int time)
  {
    delayTime = time;
    return this;
  }

  /**
   * Returns the direction from the current to the given position in degrees (-180..180).
   * Angle from north, positive clockwise.
   */
  public double towards(double x, double y)
  {
    double epsilon = 1E-12;
    double deltax = x - this.x;
    double deltay = y - this.y;
    if (Math.abs(deltay) < epsilon)
    {
      if (deltax > 0)
        return 90;
      else
        return -90;
    }

    double arc = Float11.atan2(deltax, deltay);  // atan(deltax/deltay)
    return arc / Math.PI * 180;
  }

 /**
   * Moves the turtle to the initial position at (0, 0) looking north. Other
   * properties and the traces are not changed.
   * Returns the Turtle reference.
   */
  public Turtle home()
  {
    x = 0;
    y = 0;
    phi = 0;
    _x = 0;
    _y = 0;
    shape.draw(x, y, phi, turtleColor, isVisible, isWrap);
    return this;
  }

  /**
   * Enables/disables automatic updating the window after each turtle movement.
   * If disabled, call repaint() to perform manual repainting.
   * Disabling speeds up program execution enormously.
   * Returns true, if repainting is currently enabled, otherwise false.
   * @see #repaint
   */
  public boolean enableRepaint(boolean b)
  {
    boolean oldRepaint = shape.getRepaint();
    shape.setRepaint(b);
    return oldRepaint;
  }

  /**
   * Performs manual repainting when automatic repainting is disabled.
   * Returns the Turtle reference.
   * @see #enableRepaint
   */
  public Turtle repaint()
  {
    pg.repaint();
    return this;
  }

  /**
   * Sets the color for subsequent filling.
   * Returns the Turtle reference.
   * @see #fillToPoint
   * @see #fillToHorizontal
   * @see #fillToVertical
   */
  public Turtle setFillColor(int color)
  {
    fillColor = color;
    return this;
  }

  /**
   * Sets the color in RGB-format for subsequent filling.
   * Returns the Turtle reference.
   * @see #fillToPoint
   * @see #fillToHorizontal
   * @see #fillToVertical
   */
  public Turtle setFillColor(int r, int g, int b)
  {
    int rgb = (r << 16) + (g << 8) + b;
    return setFillColor(rgb);
  }

  /**
   * Attachs a imaginary rubber band at the given position, when the turtle subsequently
   * moves around, the area that is covered by the rubber band is filled using the
   * current fill color. Filling is on until fillOff() is called.
   * Returns the Turtle reference.
   */
  public Turtle fillToPoint(double x, double y)
  {
    fillMode = FillMode.fillPoint;
    fillAnchorX = x;
    fillAnchorY = y;
    return this;
  }

  /**
   * Draws a imaginary horizonatal line at given y-coordinate,
   * when the turtle subsequently moves around, the trapezoidal area that is covered
   * by this line and the turtle position is filled using the
   * current fill color. Filling is on until fillOff() is called.
   * Returns the Turtle reference.
   */
  public Turtle fillToHorizontal(double y)
  {
    fillMode = FillMode.fillHorz;
    fillLine = y;
    return this;
  }

  /**
   * Draws a imaginary vertical line at given x-coordinate,
   * when the turtle subsequently moves around, the trapezoidal area that is covered
   * by this line and the turtle position is filled using the
   * current fill color. Filling is on until fillOff() is called.
   * Returns the Turtle reference.
   */
  public Turtle fillToVertical(double x)
  {
    fillMode = FillMode.fillVert;
    fillLine = x;
    return this;
  }

  /**
   * Stops the filling started by calling fillToPoint(), fillToHorizontal()
   * or fillToVertical().
   * Returns the Turtle reference.
   * @see #fillToPoint
   * @see #fillToHorizontal
   * @see #fillToVertical
   */
  public Turtle fillOff()
  {
    fillMode = FillMode.fillOff;
    return this;
  }

  /**
   * Gets intersection point with border.
   */
  private PointD getIntersection(double x1, double y1, double x2, double y2)
  {
//    System.out.println(" x1: " + x1 +" y1: " + y1 +" x2: " + x2 +" y2: " + y2);
    double epsilon = 1E-12;  // smallest value for denominator

    double d = x2 - x1;
    if (Math.abs(d) < epsilon)
    if (d >= 0)
      d = epsilon;
    else
      d = -epsilon;
    double gradient = (y2 - y1) / d;

    d = y2 - y1;
    if (Math.abs(d) < epsilon)
    if (d >= 0)
      d = epsilon;
    else
      d = -epsilon;
    double invgradient = (x2 - x1) / d;

    double xTop = x1 + invgradient * (range - y1);
    double xBottom = x1 + invgradient * (-range - y1);
    double yRight = y1 + gradient * (range - x1);
    double yLeft = y1 + gradient * (-range - x1);

//    System.out.println(" xTop: " + xTop + " xBottom: " + xBottom +
//                       " yRight: " + yRight +" yLeft: " + yLeft);

    // top border
    if ((y1 < range && y2 > range  // in-to-out
      || y1 > range && y2 < range) // out-to-in
      && xTop > -range && xTop < range)
      return new PointD(xTop, range, Border.top);

    // bottom border
    if ((y1 > -range && y2 < -range  // in-to-out
      || y1 < -range && y2 > -range) // out-to-in
        && xBottom > -range && xBottom < range)
      return new PointD(xBottom, -range, Border.bottom);

    // right border
    if ((x1 < range && x2 > range  // in-to-out
      || x1 > range && x2 < range) // out-to-in
      && yRight > -range && yRight < range)
      return new PointD(range, yRight, Border.right);

    // left border
    if ((x1 > -range && x2 < -range  // in-to-out
      || x1 < -range && x2 > -range) // out-to-in
      && yLeft > -range && yLeft < range)
      return new PointD(-range, yLeft, Border.left);

    return null;  // no intersecton
  }

  /**
   * Shows the playground if it's hidden by another window.
   * The turtle's playground may be hidden, because another turtle's playground
   * or some other Gui window became the current window.
   */
  public Turtle showPlayground()
  {
    pg.show(true);
    return this;
  }

  /**
   * Creates a new turtle that has exactly the same properties and traces, but
   * the given playground title.
   * Its playground is initially hidden and becomes
   * the current  window by calling showPlayground.
   * The given title is the title of the new playground.
   * Returns the reference to the new turtle.
   */
  public Turtle clone(String title)
  {
    Turtle t = new Turtle(title, bgColor, turtleColor,isVisible, false);
    t.x = x;
    t.y = y;
    t.phi = phi;
    t._x = _x;
    t._y = _y;
    t.fillAnchorX = fillAnchorX;
    t.fillAnchorY = fillAnchorY;
    t.fillLine = fillLine;
    t.penColor = penColor;
    t.fillColor = fillColor;
    t.isTraceShown = isTraceShown;
    t.isWrap = isWrap;
    t.fillMode = fillMode;
    t.delayTime = delayTime;

    // Copy parent's trace buffer to child's trace buffer
    t.pg.getTraceBuffer().drawImage(pg.getTraceImage(), 0, 0, 0);

    // Copy child's trace buffer to child's turtle buffer and add turtle image
    t.shape.draw(x, y, phi, turtleColor, isVisible, isWrap);

    return t;
  }

 /** Creates a new turtle that has exactly the same properties, playground title and traces.
  *  Its playground is initially hidden and becomes
   * the current  window by calling showPlayground.
   * Returns the reference to the new turtle.
   */
  public Turtle clone()
  {
    return clone(title);
  }

  /**
   * Draws a turtle shape of same color at the current position. This shape will
   * remain at this position when the turtle moves away.
   * Returns the Turtle reference.
   */
  public Turtle stampTurtle()
  {
    TurtleStamp ts = new TurtleStamp(x, y, phi, turtleColor);
    shape.stampedTurtles.addElement(ts);
    return this;
  }

  /**
   * Returns the turtle's playground, which is an instance of
   * javax.microedition.lcdui.Canvas.
   * Useful for low level operations, e.g. checking key events
   * with getPlayground().getGameAction(keyCode).
   */
  public Playground getPlayground()
  {
    return pg;
  }
}