// TurtleRobot.java

/*
This software is part of the NxtJLib library.
It is Open Source Free Software, so you may
- run the code for any purpose
- study how the code works and adapt it to your needs
- integrate all or parts of the code in your own programs
- redistribute copies of the code
- improve the code and release your improvements to the public
However the use of the code is entirely your responsibility.
*/

package ch.aplu.nxt;

import ch.aplu.nxt.platform.*;

/**
 * Implementation of the basic Logo turtle movements.
 */
public class TurtleRobot extends NxtRobot
{
  private String id = "";
  private Gear gear;
  private double stepFactor;
  private double rotationFactor;
  private int turtleSpeed;
  private int rotationSpeed;
  private static TurtlePane tp = null;

  /**
   * Creates a turtle instance using a NXT brick with given Bluetooth name.
   * A connection trial is engaged while an information box is shown.
   * If the connection fails, the calling thread will be put in a wait state and the program hangs.
   * The user can then terminate the application by closing the information pane.
   * established at this time.<br><br>
   * If showConsole = true, the turtle commands are displayed in a console window.
   * If btName = null, a input dialog is displayed where the Bluetooth name can be entered.
   * @param btName the Bluetooth friendly name, e.g. NXT
   * @param showPane if true, an information pane is showed, where the turtle commands are listed
   */
  public TurtleRobot(String btName, boolean showPane)
  {
    super(btName);
    id = getBtName() + ": ";
    NxtProperties props = NxtRobot.getProperties();
    stepFactor = props.getDoubleValue("TurtleStepFactor");
    rotationFactor = props.getDoubleValue("TurtleRotationFactor");
    turtleSpeed = props.getIntValue("TurtleSpeed");
    rotationSpeed = props.getIntValue("TurtleRotationSpeed");
    gear = new Gear(MotorPort.A,  MotorPort.B);
    gear.setSpeed(turtleSpeed);
    addPart(gear);
    if (showPane)
    {
      if (tp == null)  // Only one console for all TurtleRobots
        tp = new TurtlePane(this);
    }
  }

  /**
   * Same as TurtleRobot(btName, true)
   * @param btName the Bluetooth friendly name, e.g. NXT
   */
  public TurtleRobot(String btName)
  {
    this(btName, true);
  }

  /**
   * Same as TurtleRobot(btName), but ask for Bluetooth name
   */
  public TurtleRobot()
  {
    this(null, true);
  }

  /**
   * Returns the gear used as component of the turtle.
   * @return the reference of the Gear instance
   */
  public Gear getGear()
  {
    return gear;
  }

  /**
   * Sets the turtle speed to the given value.
   * The TurtleRotationSpeed remains unchanged.
   * The speed will be changed to the new value at the next movement call only.
   * @param speed 0..100
   * @return the object reference to allow method chaining
   */
  public TurtleRobot setTurtleSpeed(int speed)
  {
    if (tp != null)
      tp.println(id + "setTurtleSpeed(" + speed + ")");
    gear.setSpeed(speed);
    return this;
  }

  /**
   * Returns the current turtle speed.
   * @return speed 0..100
   */
  public int getTurtleSpeed()
  {
    return gear.getSpeed();
  }

  /**
   * Moves the turtle forward the given number of steps.
   * The methods blocks until the turtle is at the final position.
   * The actual distance depends on the "TurtleStepFactor"
   * as defined in the nxtlib.properties file.
   * The turtle speed corresponds to the "TurtleSpeed"
   * as defined in the nxtlib.properties file.
   * @param steps the number of steps to go.
   * @return the object reference to allow method chaining
   */
  public TurtleRobot forward(int steps)
  {
    if (tp != null)
      tp.println(id + "forward(" + steps + ")");
    gear.moveTo(Tools.round(stepFactor * steps));
    return this;
  }

  /**
   * Moves the turtle backward the given number of steps.
   * The methods blocks until the turtle is at the final position.
   * The actual distance depends on the "TurtleStepFactor"
   * as defined in the nxtlib.properties file.
   * The turtle speed corresponds to the "TurtleSpeed"
   * as defined in the nxtlib.properties file.
   * @param steps the number of steps to go.
   * @return the object reference to allow method chaining
   */
  public TurtleRobot backward(int steps)
  {
    if (tp != null)
      tp.println(id + "backward(" + steps + ")");
    gear.moveTo(-Tools.round(stepFactor * steps));
    return this;
  }

  /**
   * Turns the turtle to the left for the given angle.
   * The methods blocks until the turtle is at the final position.
   * Due to the construction of the NXT the accuracy is limited.
   * It depends on the "TurtleRotationFactor" and the "TurtleRotationSpeed"
   * as defined in the nxtlib.properties file.
   * @param angle the angle in degree to rotate.
   * @return the object reference to allow method chaining
   */
  public TurtleRobot left(int angle)
  {
    if (tp != null)
      tp.println(id + "left(" + angle + ")");
    int speed = gear.getSpeed();
    gear.setSpeed(rotationSpeed);
    gear.turnTo(-Tools.round(rotationFactor * angle));
    gear.setSpeed(speed);
    return this;
  }

  /**
   * Turns the turtle to the right for the given angle.
   * The methods blocks until the turtle is at the final position.
   * Due to the construction of the NXT the accuracy is limited.
   * It depends on the "TurtleRotationFactor" and the "TurtleRotationSpeed"
   * as defined in the nxtlib.properties file.
   * @param angle the angle in degree to rotate.
   * @return the object reference to allow method chaining
   */
  public TurtleRobot right(int angle)
  {
    if (tp != null)
      tp.println(id + "right(" + angle + ")");
    int speed = gear.getSpeed();
    gear.setSpeed(rotationSpeed);
    gear.turnTo(Tools.round(rotationFactor * angle));
    gear.setSpeed(speed);
    return this;
  }
 }