// NxtProperties.java, Java ME version
// Platform (Java SE, ME) dependent code
// Should be visible in package only. So not included in Javadoc

/*
This software is part of the NxtJLib library.
It is Open Source Free Software, so you may
- run the code for any purpose
- study how the code works and adapt it to your needs
- integrate all or parts of the code in your own programs
- redistribute copies of the code
- improve the code and release your improvements to the public
However the use of the code is entirely your responsibility.
 */

package ch.aplu.nxt.platform;

import java.io.*;
import ch.aplu.gidlet.*;


/**
 * Class to read properties from the NxtJLib property file.
 * Many library options are defined in a property file 'nxtjlib.properties' that
 * may be modified as needed.<br><br>
 * Consult the distributed file for more information.
 */
public class NxtProperties
{
  private String[] lines;
  private final String propFile = "nxtjlib.properties";

  /**
   * Searches the property file 'nxtjlib.properties' in subdirectory
   * 'res' and, if not found, in subdirectory 'ch/aplu/nxt/properties'
   * of the jar file. Be careful to keep the original formatting.
   */
  public NxtProperties()
  {
    // First search in subdir res
    byte[] res = Gidlet.getResource("/res/" + propFile);
    if (res == null)  // Not found
    {
      // Search in subdir ch/aplu/nxt/properties
      res = Gidlet.getResource("/ch/aplu/nxt/properties/" + propFile);
      if (res == null)
      {
        new ShowError("Properties file " + propFile + " not found");
        return;
      }
    }
    StringBuffer sb = new StringBuffer();
    ByteArrayInputStream bis = new ByteArrayInputStream(res);
    InputStreamReader isr = new InputStreamReader(bis);
    int ch;
    try
    {
      while ((ch = isr.read()) != -1)
        sb.append((char)ch);
    }
    catch (IOException ex)
    {
      System.out.println(ex);
    }
    lines = Gidlet.split(sb.toString(), "\n");
  }

  /**
   * Returns empty string. Declared for compatibility with J2SE.
   */
  public String getResourceInfo()
  {
    return "";
  }

  private String getProperty(String key)
  {
    for (int i = 0; i < lines.length; i++)
    {
     if (lines[i].length() > 0 && lines[i].charAt(0) != '#')
     {
        String[] property = Gidlet.split(lines[i], "=");
        if (property.length == 2 && property[0].trim().equals(key))
          return property[1].trim();
      }
    }
    return null;
  }

  /**
   * Gets the property value with the given key as string.
   * An error form is shown, if the property does not exist.
   * @param key the key of the property
   * @return the value of the property or empty string, if the property is  not found
   */
  public String getStringValue(String key)
  {
    String value = getProperty(key);
    if (value == null)
    {
      new ShowError("Property '" + key +
        "' not found in file '" + propFile + "'");
      return "";
    }
    return value;
  }

  /**
   * Gets the property value with the given key as int.
   * An error form is shown, if the property does not exist or has bad format.
   * @param key the key of the property
   * @return the value of the property or 0,
   * if the property is not found or does not have integer format
   */
  public int getIntValue(String key)
  {
    String value = getProperty(key);
    if (value == null)
    {
      new ShowError("Property '" + key +
        "' not found in file '" + propFile + "'");
      return 0;
    }
    int intValue;
    try
    {
      intValue = Integer.parseInt(value);
    }
    catch (NumberFormatException ex)
    {
      new ShowError("Illegal integer format of property '" + key +
        "' in file '" + propFile + "'");
      return 0;
    }
    return intValue;
  }

  /**
   * Gets the property value with the given key as double.
   * An error form is shown, if the property does not exist or has bad format.
   * @param key the key of the property
   * @return the value of the property or 0,
   * if the property is not found or does not have double format
   */
  public double getDoubleValue(String key)
  {
    String value = getProperty(key);
    if (value == null)
    {
      new ShowError("Property '" + key +
        "' not found in file '" + propFile + "'");
      return 0;
    }
    double doubleValue;
    try
    {
      doubleValue = Double.parseDouble(value);
    }
    catch (NumberFormatException ex)
    {
      new ShowError("Illegal double format of property '" + key +
        "' in file '" + propFile + "'");
      return 0;
    }
    return doubleValue;
  }
}

